# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring

import os

PKG_DIR = os.path.abspath(os.path.dirname(__file__))


def pillar_dirs():
    return [os.path.join(PKG_DIR, "pillar")]


def renderers_dirs():
    return [os.path.join(PKG_DIR, "renderers")]



# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring
# pylint: disable=redefined-outer-name

import os

import pytest
import salt
from salt.exceptions import SaltRenderError


def test_common_wildcard(env):
    env.setup(
        {
            "tower.sls": """
            base:
                - common/*
            """,
            "common/test.sls": """
            common: True
            """,
        }
    )

    assert env.ext_pillar() == {"common": True}


def test_missing_file(env):
    env.setup(
        {
            "tower.sls": """
            base:
                - common.sls
            """,
        }
    )

    assert env.ext_pillar() == {}


def test_missing_file_raises(env):
    env.opts.update({"salt_tower.raise_on_missing_files": True})
    env.setup(
        {
            "tower.sls": """
            base:
                - common.sls
            """,
        }
    )

    with pytest.raises(FileNotFoundError, match=r"common.sls"):
        env.ext_pillar()


def test_missing_file_raises_wildcard(env):
    env.opts.update({"salt_tower.raise_on_missing_files": True})
    env.setup(
        {
            "tower.sls": """
            base:
                - common/*
            """,
        }
    )

    with pytest.raises(FileNotFoundError, match=r"common/\*"):
        env.ext_pillar()


def test_missing_file_raises_wildcard_present(env):
    env.opts.update({"salt_tower.raise_on_missing_files": True})
    env.setup(
        {
            "tower.sls": """
            base:
                - common/*
            """,
            "common/file.sls": """
            key: 1
            """,
        }
    )

    assert env.ext_pillar() == {"key": 1}


def test_missing_file_raises_optional(env):
    env.opts.update({"salt_tower.raise_on_missing_files": True})
    env.setup(
        {
            "tower.sls": """
            base:
                - !optional common.sls
                - "minion":
                    - !optional common.sls
            """,
        }
    )

    assert env.ext_pillar() == {}


def test_missing_file_raises_wildcard_optional(env):
    env.opts.update({"salt_tower.raise_on_missing_files": True})
    env.setup(
        {
            "tower.sls": """
            base:
                - !optional common/*
                - "minion":
                    - !optional common/*
            """,
        }
    )

    assert env.ext_pillar() == {}


def test_jinja(env):
    env.setup(
        {
            "tower.sls": """
            base:
                - {{ grains['id'] }}.sls
            """,
            "test_master.sls": """
            minion:
                id: {{ grains['id'] }}
            """,
        }
    )

    assert env.ext_pillar() == {"minion": {"id": "test_master"}}


def test_match_minion_id(env):
    env.setup(
        {
            "tower.sls": """
            base:
                - test_*:
                    - match: minion_id
            """
        }
    )

    assert env.ext_pillar() == {"match": "minion_id"}


def test_match_extended_regex(env):
    env.setup(
        {
            "tower.sls": """
            base:
                - E@[tes]{4}_master$:
                    - match: regexp
            """
        }
    )

    assert env.ext_pillar() == {"match": "regexp"}


def test_match_pillar_data(env):
    env.pillar.update({"pillar_key": True})
    env.setup(
        {
            "tower.sls": """
            base:
                - I@pillar_key:True:
                    - match: pillar
            """
        }
    )

    assert env.ext_pillar() == {
        "pillar_key": True,
        "match": "pillar",
    }


def test_match_tower_pillar_key(env):
    env.setup(
        {
            "tower.sls": """
            base:
                - '*':
                    - minion: {id: 'value'}
                - I@minion:id:*:
                    - match: tower_pillar
            """
        }
    )

    assert env.ext_pillar() == {"minion": {"id": "value"}, "match": "tower_pillar"}


def test_match_grains(env):
    env.grains.update({"virtual": False})
    env.setup(
        {
            "tower.sls": """
            base:
                - G@virtual:*:
                    - match: grains
            """
        }
    )

    assert env.ext_pillar() == {"match": "grains"}


@pytest.mark.skip(reason="Global format currently disable as it's a high-risk feature")
def test_late_bind(env):
    env.setup(
        {
            "tower.sls": """
            base:
                - '*':
                    - minion: {id: 'my-minion-id'}
                    - late_bind: {minion_id: '{minion.id}'}
            """
        }
    )

    assert env.ext_pillar() == {
        "minion": {"id": "my-minion-id"},
        "late_bind": {"minion_id": "my-minion-id"},
    }


def test_late_bind_not_matching(env):
    env.setup(
        {
            "tower.sls": """
            base:
                - '*':
                    - late_bind: '{fuubar}'
            """
        }
    )

    assert env.ext_pillar() == {"late_bind": "{fuubar}"}


def test_late_bind_invalid(env):
    env.setup(
        {
            "tower.sls": """
            base:
                - '*':
                    - late_bind: "{^$://DJF$$}"
            """
        }
    )

    assert env.ext_pillar() == {"late_bind": "{^$://DJF$$}"}


def test_default_renderers(env):
    env.setup(
        {
            "tower.sls": """
            base:
                - pillar.sls
            """,
            "pillar.sls": """
            yaml: 'yes'
            jinja: '{{ "yes" }}'
            """,
        }
    )

    assert env.ext_pillar() == {"yaml": "yes", "jinja": "yes"}


def test_renderers_shebang_py(env):
    env.setup(
        {
            "tower.sls": """
            base:
                - pillar.sls
            """,
            "pillar.sls": """
            #!py

            def run():
                return {'py': 'yes'}
            """,
        }
    )

    assert env.ext_pillar() == {"py": "yes"}


def test_include(env):
    env.setup(
        {
            "tower.sls": """
            base:
                - common
            """,
            "common.sls": """
            include:
                - else.sls
            """,
            "else.sls": """
            else: True
            """,
        }
    )

    assert env.ext_pillar() == {"else": True}


def test_include_relative(env):
    env.setup(
        {
            "tower.sls": """
            base:
                - path/to/file
            """,
            "path/to/file.sls": """
            include:
                - ./else.sls
            """,
            "path/to/else.sls": """
            else: True
            """,
        }
    )

    assert env.ext_pillar() == {"else": True}


def test_include_relative_up(env):
    env.setup(
        {
            "tower.sls": """
            base:
                - path/to/file
            """,
            "path/to/file.sls": """
            include:
                - ../else.sls
            """,
            "path/else.sls": """
            else: True
            """,
        }
    )

    assert env.ext_pillar() == {"else": True}


def test_include_nested(env):
    env.setup(
        {
            "tower.sls": """
            base:
                - a/common
            """,
            "a/common.sls": """
            include:
                - b/file.sls
            """,
            "b/file.sls": """
            test: True
            """,
        }
    )

    assert env.ext_pillar() == {"test": True}


def test_include_nested_init(env):
    env.setup(
        {
            "tower.sls": """
            base:
                - a/common
            """,
            "a/common.sls": """
            include:
                - b
            """,
            "b/init.sls": """
            test: True
            """,
        }
    )

    assert env.ext_pillar() == {"test": True}


def test_include_nested_glob(env):
    env.setup(
        {
            "tower.sls": """
            base:
                - a/common
            """,
            "a/common.sls": """
            include:
                - b/*.sls
            """,
            "b/one.sls": """
            one: True
            """,
            "b/two.sls": """
            two: True
            """,
        }
    )

    assert env.ext_pillar() == {"one": True, "two": True}


def test_include_recursive(env):
    env.setup(
        {
            "tower.sls": """
            base:
                - common
            """,
            "common.sls": """
            include: [else.sls]
            order: [1]
            """,
            "else.sls": """
            include: [common.sls]
            order: [2]
            """,
        }
    )

    assert env.ext_pillar() == {"order": [2, 1]}


def test_context_tmpl(env, tmpdir):
    env.setup(
        {
            "tower.sls": """
            base:
                - template.sls
            """,
            "template.sls": """
            tmplpath: {{ tmplpath }}
            tmpldir: {{ tmpldir }}
            """,
        }
    )

    assert env.ext_pillar() == {
        "tmplpath": os.path.join(tmpdir, "template.sls"),
        "tmpldir": tmpdir,
    }


def test_context_minion_id(env):
    env.setup(
        {
            "tower.sls": """
            base:
                - template.sls
            """,
            "template.sls": """
            minion_id: {{ minion_id }}
            """,
        }
    )

    assert env.ext_pillar() == {"minion_id": "test_master"}


def test_context_basedir(env):
    """
    Test path lookup behavior of the yamlet renderer in a tower.

    The !include tag should lookup non-relative paths from the tower
    root.
    """
    env.setup(
        {
            "tower.sls": """
            base:
                - common/init.sls
            """,
            "common/init.sls": """
            conf: !include {{ basedir }}/files/conf.txt
            """,
            "files/conf.txt": """
            Some config file...
            """,
        }
    )

    assert env.ext_pillar() == {"conf": "Some config file..."}


def test_error_invalid_file(env):
    """
    Rendering errors shall not be ignored but raised. Otherwise an
    incomplete or wrong pillar dataset might be returned and the error
    would go unnoticed.
    """
    env.setup(
        {
            "tower.sls": """
            base:
                - common/init.sls
            """,
            "common/init.sls": """
            {% abc %}
            """,
        }
    )

    with pytest.raises(SaltRenderError):
        env.ext_pillar()


def test_render_func(env):
    env.setup(
        {
            "tower.sls": """
            base:
                - index
            """,
            "index.sls": """
            key: !include test/conf.j2
            """,
            "test/conf.j2": """
            #! jinja | text strip
            {{ render('test/conf2.j2') }}
            """,
            "test/conf2.j2": """
            This is a test.
            """,
        }
    )

    assert env.ext_pillar() == {"key": "This is a test."}


def test_render_func_relative(env):
    """
    Relative paths must be specified using the `tmpldir` variable.
    """
    env.setup(
        {
            "tower.sls": """
            base:
                - index
            """,
            "index.sls": """
            key: !include test/conf.j2
            """,
            "test/conf.j2": """
            #! jinja | text strip
            {{ render(tmpldir + '/conf2.j2') }}
            """,
            "test/conf2.j2": """
            This is a test.
            """,
        }
    )

    assert env.ext_pillar() == {"key": "This is a test."}


def test_render_func_in_top(env):
    env.setup(
        {
            "tower.sls": """
            base:
                - '*':
                    - key: !include test/conf.j2
            """,
            "test/conf.j2": """
            #! jinja | text strip
            {{ render(tmpldir + '/conf2.j2') }}
            """,
            "test/conf2.j2": """
            This is a test.
            """,
        }
    )

    assert env.ext_pillar() == {"key": "This is a test."}


def test_render_func_context(env):
    env.setup(
        {
            "tower.sls": """
            base:
                - '*':
                    - key: !include context/test.py
            """,
            "context/test.py": """
            #!py
            import os.path
            def run():
                return context["render"](
                    os.path.join(context["tmplpath"], "../template.j2"),
                    context={**context, "message": "Hello World!"}
                )
            """,
            "context/template.j2": """
            #!jinja|text strip
            {{ message }}
            """,
        }
    )

    assert env.ext_pillar() == {"key": "Hello World!"}


def test_render_include_dir_mode_sls(env):
    env.opts.update({"salt_tower.include_directory_mode": "all-sls"})
    env.setup(
        {
            "tower.sls": """
            base:
                - '*':
                    - role/openssh
            """,
            "role/openssh/10-config.sls": """
            openssh:
                config: 1
            """,
            "role/openssh/20-extension.sls": """
            openssh:
                extension: 1
            """,
        }
    )

    assert env.ext_pillar() == {"openssh": {"config": 1, "extension": 1}}


def test_tower_jinja_import(env):
    """
    The JINJA renderer will always lookup relative files (e.g. from
    import or include) using either the states roots or the pillar roots
    directories if a `saltenv` is set (which is the case when
    renderering pillars).

    Unfortunately it is *not* possible to pass any customization to the
    renderer, e.g. passing an own function to resolve files therefore it
    currently is impossible to load files from the tower directory
    *unless* it also is a pillars root directory.

    This tests ensures that the renderers are called *without* a saltenv
    and do lookup files relative to the current template file.
    """
    env.setup(
        {
            "tower.sls": """
            base:
                - '*':
                    - context/init.sls
            """,
            "context/init.sls": """
            #!jinja|yaml
            {%- import_yaml "data.yaml" as data %}
            foo: {{ data.key }}
            """,
            "context/data.yaml": """
            key: 1
            """,
        }
    )

    assert env.ext_pillar() == {"foo": 1}


@pytest.mark.skipif(salt.version.__version__ < "2018", reason="requires salt 2018+")
def test_tower_jinja_import_roots(env):
    """
    Test that the rendering engines are called with all flags to import
    files from the pillar roots instead of the states root. This
    behavior is disabled by default, but can be enabled by a config
    flag.

    Some rendering engines will behave differently now and lookup files
    in the salt masters pillar_roots instead of realtive to the current
    file. The exact behavor depends on salt internals and cannot be
    customized.
    """
    env.opts.update({"salt_tower.unstable_enable_saltenv": True})
    env.setup(
        {
            "tower.sls": """
            base:
                - '*':
                    - context/init.sls
            """,
            "context/init.sls": """
            #!jinja|yaml
            {%- import_yaml "data.yaml" as data %}
            foo: {{ data.source }}
            """,
        }
    )

    assert env.ext_pillar() == {"foo": "pillar_root"}



# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring
# pylint: disable=redefined-outer-name


def test_render(render):
    template = """
        #!yaml | filter grain=id
        test_*:
          key: 1

        test_master:
          key: 2

        something else:
          key: 3
    """

    assert render(template) == {"key": 1}


def test_render_default_minion_id(render):
    template = """
        #!yaml | filter
        test_master:
          key: 1

        something else:
          key: 2
    """

    assert render(template) == {"key": 1}


def test_render_grain(env, render):
    env.grains.update({"os_family": "Debian"})

    template = """
        #!yaml | filter grain=os_family
        Debian:
          key: 1

        something else:
          key: 2
    """

    assert render(template) == {"key": 1}


def test_render_pillar(render):
    template = """
        #!yaml | filter pillar=some:key
        value:
          match: True
    """

    context = {"pillar": {"some": {"key": "value"}}}

    assert render(template, context=context) == {"match": True}


def test_render_default(render):
    template = """
        #!yaml | filter pillar=some:key default='matching value'
        '*value':
          match: True
    """

    assert render(template) == {"match": True}


def test_render_key(render):
    template = """
        #!yaml | filter grain=id key=sub:path
        'test_master':
          key: 2
    """

    assert render(template) == {"sub": {"path": {"key": 2}}}



# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring
# pylint: disable=redefined-outer-name
# pylint: disable=unidiomatic-typecheck

import os
import textwrap

from salt.utils.odict import OrderedDict


def test_render(render):
    template = """
      base:
        key: 'text'
        list: [1, 2]
    """

    assert render(template, default="yamlet") == {
        "base": {"key": "text", "list": [1, 2]}
    }


def test_ordereddict(render):
    template = """
      base:
        key: 'text'
        list: [1, 2]
    """

    result = render(template, default="yamlet")
    assert type(result) == OrderedDict


def test_tag_read(env, render):
    env.write("other_file.txt", "content")

    template = """
      key: !read other_file.txt
    """

    assert render(template, default="yamlet") == {"key": b"content"}


def test_tag_read_binary(env, render):
    env.write("binary-file.bin", b"jd\x81\xed\xa2~*\xca6\xd88,\x15zr\xb6", mode="wb")

    template = """
      key: !read binary-file.bin
    """

    assert render(template, default="yamlet") == {
        "key": b"jd\x81\xed\xa2~*\xca6\xd88,\x15zr\xb6"
    }


def test_tag_include(env, render):
    env.write("template.sls", "another: value")

    assert render("key: !include template.sls", default="yamlet") == {
        "key": {"another": "value"}
    }


def test_include_yamlet_jinja(env, render):
    env.write(
        "some_yamlet.sls",
        """
            read: !read file.txt
            id: {{ grains["id"] }}
        """,
    )
    env.write("file.txt", "txt")

    template = """
        key: !include some_yamlet.sls
    """

    assert render(template, default="yamlet") == {
        "key": {"read": b"txt", "id": "test_master"}
    }


def test_include_mapping(env, render):
    env.write("template.sls", "another: value")

    template = """
        key: !include
            source: template.sls
    """

    assert render(template, default="yamlet") == {"key": {"another": "value"}}


def test_include_mapping_context(env, render):
    env.write("template.sls", "val: {{ value }}")

    template = """
        key: !include
          source: template.sls
          context:
            value: 5
    """

    assert render(template, default="yamlet") == {"key": {"val": 5}}


def test_include_shebang(env, render):
    """
    The !include tag must respect a shebang in the given template.

    It itself does not strip the shebang line (use the text renderer) for that
    but it must handle IO-like objects such as returned by the jinja renderer.
    """

    env.write("template.sls", "#!jinja\nval: {{ grains['id'] }}")

    template = """
        key: !include template.sls
    """

    assert render(template, default="yamlet") == {"key": "#!jinja\nval: test_master"}


def test_include_tmplpath(env, render, tmpdir):
    """
    The !include tag passes ``tmplpath`` and ``tmpldir`` to the template
    renderer to allow relative includes via ``tmpldir + '/file'``.
    """

    env.write(
        "file.sls",
        textwrap.dedent(
            """
            #!jinja|yaml
            tmplpath: {{ tmplpath }}
            tmpldir: {{ tmpldir }}
            """
        ).strip(),
    )

    template = """
      key: !include file.sls
    """

    assert render(template, default="yamlet") == {
        "key": {"tmplpath": os.path.join(tmpdir, "file.sls"), "tmpldir": tmpdir}
    }


def test_include_inherit_context(env, render):
    """
    The !include tag passes given ``context`` to downstream renderers.
    """

    env.write(
        "file.sls",
        textwrap.dedent(
            """
            #!jinja|text
            {{ fuubar() }}
            """
        ).strip(),
    )

    template = """
      key: !include file.sls
    """

    def fuubar():
        return "fuubar"

    assert render(template, default="yamlet", context={"fuubar": fuubar}) == {
        "key": "fuubar"
    }


def test_include_does_not_mutate_context(env, render):
    """
    The !include tag passes given ``context`` to downstream renderers.

    They usual mutate their ``context`` object, but yamlet passes a copy.
    Therefore the original ``context`` object is not modified.
    """

    env.write("file.sls", "")

    context = {"key": "abc"}
    template = """
      key: !include file.sls
    """

    render(template, default="yamlet", context=context)

    assert context == {"key": "abc"}



# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring
"""
A text renderer for processing and loading text blobs.

It is mainly developed to be used with file loaded in tower pillars and
to be used with the yamlet renderers include tag:

.. code-block:: yaml

    pillar:
      key: !include ./file.txt

.. code-block:: text

    #!text

    The pillar key content.


The following example will render a text file and strip whitespace from
both ends:

.. code-block:: text

    #!text strip

    A short text message


The renderer is further able to wrap the text blob into a nested
dictionary:

.. code-block:: text

    #!text strip key=path:to:blob

    A short text message


This will return the following python dict:

.. code-block:: python

    {'path': {'to': {'blob': 'A short text message'}}}

"""


def render(blob, _saltenv, _sls, argline=None, key=None, **_kwargs):
    if not isinstance(blob, str):
        blob = blob.read()

    if blob.startswith("#!"):
        blob = blob[(blob.find("\n") + 1) :]

    if argline is not None:
        for arg in argline.split(None):
            if "=" in arg:
                k, v = arg.split("=", 1)
            else:
                k, v = arg, None

            if k == "strip":
                blob = blob.strip()
            elif k == "key":
                key = v

    if key:
        for k in reversed(key.split(":")):
            blob = {k: blob}

    return blob



# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring
"""
The ``filter`` renderers takes a dataset and only returns the first
matching value. It can match globs on grains and pillar values.

The input must be a parsed dictionary, for example from the YAML
renderer. The first key is uses as a pattern to match the value from
grains or pillar. Shell like globs are supported as the ``fnmatch``
function is used to check the patterns.

The ``default`` option can provide a string used as the value if the
grain or pillar key does not exist.

Example: Default matching uses ``minion_id``:

.. code-block:: yaml

    #!yaml | filter

    minion-1:
      some:
        data: for minion-1

    other-minion:
      some:
        data: for other minion

    minions*:
      some:
        data: for mulitple minions matching key


Example: Matching using the ``os_family`` grain

.. code-block:: yaml

    #!yaml | filter grain=os_family default='default value'

    Debian:
      package_name: docker.io

    default value:
      package_name: docker-ce
"""

import logging
import shlex
from fnmatch import fnmatch

try:
    from salt.exceptions import TemplateError
except ImportError:
    from salt.exceptions import SaltException as TemplateError

try:
    from salt.utils.data import traverse_dict_and_list
except ImportError:
    from salt.utils import traverse_dict_and_list


VALID_SELECTORS = ("grain", "pillar")
LOG = logging.getLogger(__name__)


def render(  # pylint: disable=too-many-branches
    source, _saltenv, _sls, argline=None, **kwargs
):
    if not isinstance(source, dict):
        raise TypeError(f"Source must be a dict, not {type(source)}")

    selector = "grain"
    default = None
    lookup_key = "id"
    wrap_key = False

    if argline:
        for arg in shlex.split(argline):
            try:
                (option, value) = arg.split("=", 2)
            except ValueError:
                option, value = arg, None

            if option in VALID_SELECTORS:
                if not value:
                    raise TemplateError(f"Selector {option!r} needs a value")
                selector = option
                lookup_key = value

            elif option == "default":
                if not value:
                    raise TemplateError(f"Option {option!r} needs a value")
                default = value

            elif option == "key":
                if not value:
                    raise TemplateError(f"Option {option!r} needs a value")
                wrap_key = value

            else:
                raise TemplateError(f"Unknown option {option!r}")

    if selector == "grain":
        value = traverse_dict_and_list(__grains__, lookup_key, default)

    elif selector == "pillar":
        context = kwargs.get("context", {})
        if "pillar" in context:
            value = traverse_dict_and_list(context["pillar"], lookup_key, default)
        else:
            value = traverse_dict_and_list(__pillar__, lookup_key, default)

    if not value:
        LOG.debug("Skipping blank filter value: %r", value)
        return {}

    # Matching only works on strings
    value = str(value)

    result = {}
    for pattern in source:
        if fnmatch(value, pattern):
            result = source[pattern]
            break
    else:
        LOG.debug("No pattern matched value: %r", value)

    if wrap_key:
        for k in reversed(wrap_key.split(":")):
            result = {k: result}

    return result



# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring
# pylint: disable=redefined-outer-name

import inspect

import pytest


@pytest.fixture
def mod(env):
    return inspect.getmodule(env.build().pillars["tower"])


@pytest.fixture
def tower(mod):
    return mod.Tower(
        "minion-id",
        "base",
        {
            "pillar": {
                "key": "value",
                "list": [1, 2],
            }
        },
    )


def test_get_traverse(tower):
    assert tower.get("pillar:key") == "value"


def test_get_default(tower):
    assert tower.get("pillar:none") is None


def test_get_default_value(tower):
    assert tower.get("pillar:none", default=5) == 5


def test_get_default_require(tower):
    with pytest.raises(KeyError):
        tower.get("pillar:none", require=True)


def test_update(tower):
    tower.update(
        {
            "pillar": {
                "key": "new_value",
                "list": [3],
            }
        }
    )

    assert tower == {
        "pillar": {
            "key": "new_value",
            "list": [1, 2, 3],
        }
    }


def test_update_override(tower):
    """
    Merging different types will override
    """
    tower.update({"pillar": {"list": "fake"}})

    assert tower["pillar"]["list"] == "fake"


def test_merge(tower):
    tgt = {"a": True, "b": [1]}
    ret = tower.merge(tgt, {"b": [2]})

    assert ret is tgt
    assert ret == {"a": True, "b": [1, 2]}


def test_merge_none(tower):
    tgt = {"a": {"b": 1}}
    ret = tower.merge(tgt, {"a": None})

    assert ret == {"a": None}


def test_merge_copy(tower):
    tgt = {}
    mod = {"a": {"b": 2}}

    tower.merge(tgt, mod)

    assert tgt == {"a": {"b": 2}}
    assert tgt["a"] is not mod["a"]


def test_merge_copy_list(tower):
    tgt = [{"a": 1}]
    mod = [{"b": 2}]

    tower.merge(tgt, mod)

    assert tgt == [{"a": 1}, {"b": 2}]
    assert tgt[1] is not mod[0]


def test_merge_list_strategy_remove(tower):
    tgt = ["a", "b"]
    mod = [{"__": "remove"}, "a"]

    tower.merge(tgt, mod)

    assert tgt == ["b"]


def test_merge_list_strategy_remove_non_existant(tower):
    tgt = ["a", "b"]
    mod = [{"__": "remove"}, "c"]

    tower.merge(tgt, mod)

    assert tgt == ["a", "b"]


def test_merge_list_strategy_merge_first(tower):
    tgt = ["a", "b"]
    mod = [{"__": "merge-first"}, "c"]

    tower.merge(tgt, mod)

    assert tgt == ["c", "a", "b"]


def test_merge_list_strategy_merge_overwrite(tower):
    tgt = ["a", "b"]
    mod = [{"__": "overwrite"}, "c"]

    tower.merge(tgt, mod)

    assert tgt == ["c"]


def test_merge_list_remove_nested_doubledash(tower):
    tgt = {"a": 0}
    mod = {"a": [{"__": "overwrite"}, "a", "b"]}

    tower.merge(tgt, mod)

    assert tgt == {"a": ["a", "b"]}


def test_merge_list_remove_nested_doubledash_list(tower):
    tgt = [0]
    mod = [[{"__": "overwrite"}, 1]]

    tower.merge(tgt, mod)

    assert tgt == [0, [1]]


def test_merge_dict_strategy_remove(tower):
    tgt = {"a": 0, "b": 1}
    mod = {"__": "remove", "a": None}

    tower.merge(tgt, mod)

    assert tgt == {"b": 1}


def test_merge_dict_strategy_remove_non_existant(tower):
    tgt = {"a": 0, "b": 1}
    mod = {"__": "remove", "c": None}

    tower.merge(tgt, mod)

    assert tgt == {"a": 0, "b": 1}


def test_merge_dict_strategy_merge_first(tower):
    tgt = {"a": 0, "b": 1, "d": [4]}
    mod = {"__": "merge-first", "a": 1, "c": 2, "d": [5]}

    tower.merge(tgt, mod)

    assert tgt == {"a": 0, "b": 1, "c": 2, "d": [5, 4]}


def test_merge_dict_strategy_merge_overwrite(tower):
    tgt = {"a": 0, "b": 1}
    mod = {"__": "overwrite", "c": 2}

    tower.merge(tgt, mod)

    assert tgt == {"c": 2}


def test_merge_dict_remove_nested_doubledash(tower):
    tgt = {"a": 0}
    mod = {"a": {"__": "overwrite", "key": 1}}

    tower.merge(tgt, mod)

    assert tgt == {"a": {"key": 1}}


def test_format(tower):
    tower.update({"app": {"name": "MyApp"}})

    assert tower.format("X{app.name}X") == "XMyAppX"


def test_format_bytes(tower):
    tower.update({"app": {"name": "MyApp"}})

    assert tower.format(b"X{app.name}X") == b"X{app.name}X"


def test_format_dict(tower):
    tower.update({"ports": [80, 443]})

    assert tower.format({"bind": "0.0.0.0:{ports.0}"}) == {"bind": "0.0.0.0:80"}


def test_format_list(tower):
    tower.update({"ports": [80, 443]})

    assert tower.format(["{ports.0}", "{ports.1}"]) == ["80", "443"]


def test_format_arg(tower):
    assert tower.format({"app": "{name}"}, name="APP") == {"app": "APP"}



# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring
"""
salt-tower ext_pillar module
"""
from __future__ import absolute_import

import copy
import errno
import logging
import os
import string
from glob import glob
from typing import TYPE_CHECKING, Any, Dict, NamedTuple

import salt.loader
import salt.minion
import salt.template
import salt.utils.context
from yaml.constructor import ConstructorError
from yaml.nodes import ScalarNode

try:
    from salt.utils.data import traverse_dict_and_list
except ImportError:
    from salt.utils import traverse_dict_and_list

LOGGER = logging.getLogger(__name__)


if TYPE_CHECKING:
    __grains__: salt.utils.context.NamespacedDictWrapper
    __opts__: Dict[str, Any]
    __salt__: salt.loader.LazyLoader


if hasattr(salt.loader, "matchers"):
    # Available since salt 2019.2
    def _match_minion_impl(tgt, opts):
        matchers = salt.loader.matchers(  # pylint: disable=no-member
            dict(__opts__, **opts)
        )
        return matchers["compound_match.match"](tgt)

else:

    def _match_minion_impl(tgt, opts):
        return salt.minion.Matcher(  # pylint: disable=no-member
            opts, __salt__
        ).compound_match(tgt)


def ext_pillar(minion_id, pillar, *args, **_kwargs):
    env = __opts__.get("environment", None)

    if env is None:
        env = "base"

    tower = Tower(minion_id, env, pillar)

    for top in list(args):
        if not os.path.exists(top):
            LOGGER.warning("Tower top file `%s' does not exists", top)
            continue

        tower.run(top)

    return dict(tower)


class Optional(NamedTuple):
    pattern: str


def _yamlet_construct_optional(_loader, node):
    if isinstance(node, ScalarNode):
        return Optional(node.value)

    raise ConstructorError(  # pylint: disable=duplicate-code
        None,
        None,
        f"expected a scalar node, but found {node.id}",
        node.start_mark,
    )


class Tower(dict):
    def __init__(self, minion_id, env, pillar):
        super().__init__(pillar)

        self.env = env
        self.minion_id = minion_id

        opts = copy.copy(__opts__)
        opts["pillar"] = self

        self._renderers = salt.loader.render(opts, __salt__)
        self._formatter = Formatter(self)
        self._included = []

        if "yamlet" in self._renderers:
            self._default_renderers = "jinja|yamlet"
        else:
            LOGGER.warning(
                "Yamlet renderer not available. Tower functionality will be limited."
            )
            self._default_renderers = "jinja|yaml"

    def get(self, key, default=None, require=False, **kwargs):
        if require:
            default = KeyError(f"Pillar key missing: {key}")
        value = traverse_dict_and_list(self, key, default, **kwargs)
        if isinstance(value, KeyError):
            raise value  # pylint: disable=raising-non-exception
        return value

    def update(self, obj, merge=True, **kwargs):
        if merge:
            return self.merge(self, obj, **kwargs)
        return super().update(obj)

    def merge(self, *args, **kwargs):  # pylint: disable=no-self-use
        return _merge(*args, **kwargs)

    def format(self, obj, *args, **kwargs):
        if isinstance(obj, dict):
            return {k: self.format(v, *args, **kwargs) for k, v in obj.items()}

        if isinstance(obj, list):
            return [self.format(i, *args, **kwargs) for i in obj]

        if isinstance(obj, bytes):
            return obj

        if isinstance(obj, str):
            try:
                return self._formatter.format(obj, *args, **kwargs)
            except ValueError:
                return obj

        return obj

    def run(self, top):
        LOGGER.debug("Process tower top file `%s'", top)

        base = os.path.dirname(top)

        for item in self._load_top(top, base):
            if isinstance(item, dict):
                for tgt, items in item.items():
                    if not isinstance(items, list):
                        raise ValueError(
                            f"tower.sls: include items must be a list but is {type(items).__name__}"
                        )

                    if not self._match_minion(tgt):
                        continue

                    for itm in items:
                        self._load_item(base, itm)
            else:
                self._load_item(base, item)

    def _match_minion(self, tgt):
        try:
            if hasattr(__grains__, "value"):
                grains = __grains__.value()
            else:
                grains = __grains__
            return _match_minion_impl(
                tgt, {"grains": grains, "pillar": dict(self), "id": self.minion_id}
            )
        except Exception as err:  # pylint: disable=broad-except
            LOGGER.exception(err)
            return False

    def _load_top(self, top, base):
        data = self._compile(
            top,
            context={"basedir": base},
            _yamlet_constructors={"!optional": _yamlet_construct_optional},
        )

        if not isinstance(data, dict):
            LOGGER.critical("Tower top must be a dict, but is %s.", type(data))
            return []

        if self.env not in data:
            LOGGER.warning(
                "Tower top `%s' does not include env %s, skipping.", top, self.env
            )
            return []

        if not isinstance(data[self.env], list):
            LOGGER.critical(
                "Tower top `%s' env %s must be a list, but is %s.",
                top,
                self.env,
                type(data[self.env]),
            )
            return []

        return data[self.env]

    def _load_item(self, base, item):
        if isinstance(item, dict):
            self.update(item, merge=True)
        elif isinstance(item, str):
            self.load(item, base)
        elif isinstance(item, Optional):
            self.load(item.pattern, base, optional=True)
        else:
            raise ValueError(f"Invalid item type: {type(item).__name__}")

    def lookup(self, item, base=None, cwd=None):
        if cwd and (item.startswith("./") or item.startswith("../")):
            path = os.path.join(cwd, self.format(item))
        elif base:
            path = os.path.join(base, self.format(item))
        else:
            path = self.format(item)

        match = glob(path)

        if match:
            match = [i for i in match if os.path.isfile(i)]

        if match:
            LOGGER.debug("Found glob match: %s", match)
            return sorted(match)

        mode = __opts__.get("salt_tower.include_directory_mode")
        if mode == "all-sls":
            # This mode first checks if the given path is a directory. If yes,
            # all `*.sls` files inside the directory will be returned, ordered
            # by filename.
            if os.path.isdir(path):
                LOGGER.debug("Found directory match: %s", path)
                return sorted(
                    [
                        file
                        for file in glob(os.path.join(path, "*.sls"))
                        if os.path.isfile(file)
                    ]
                )

        for match in [path, f"{path}.sls", f"{path}/init.sls"]:
            if os.path.isfile(match):
                LOGGER.debug("Found file match: %s", match)
                return [match]

        return []

    def load(self, item, base=None, cwd=None, optional=False):
        matches = self.lookup(item, base, cwd)

        if (
            not matches
            and not optional
            and __opts__.get("salt_tower.raise_on_missing_files")
        ):
            raise FileNotFoundError(f"File not found: {item}")

        for file in matches:
            self._load_file(file, base)

    def _load_file(self, file, base=None):
        """
        Try to load fully qualified file path as a tower data file.

        It will first check if the file has already been loaded and, if not,
        if it exist in the filesystem.

        The file will be compiled and run as a salt template and is expected
        to return a dict.

        If the dict includes an `include` item or list, these includes will
        loaded before the loaded data is merged into the current pillar.
        """
        if file in self._included:
            LOGGER.warning("Skipping already included file: %s", file)
            return

        if not os.path.isfile(file):
            LOGGER.warning("Skipping non-existing file: %s", file)
            return

        self._included.append(file)

        data = self._compile(file, context={"basedir": base})

        if not isinstance(data, dict):
            LOGGER.debug("Loading %s did not return dict, but %s", file, type(data))
            return

        if "include" in data:
            includes = data.pop("include")

            if not isinstance(includes, list):
                includes = [includes]

            for include in includes:
                self.load(include, base, cwd=os.path.dirname(file))

        self.update(data, merge=True)

    def _compile(  # pylint: disable=too-many-arguments
        self,
        template,
        default=None,
        blacklist=None,
        whitelist=None,
        context=None,
        **kwargs,
    ):
        if default is None:
            default = self._default_renderers

        ctx = {}

        if context:
            ctx.update(context)

        ctx["tmplpath"] = template
        ctx["tmpldir"] = os.path.dirname(template)
        ctx["minion_id"] = self.minion_id
        ctx["pillar"] = self
        ctx["tower"] = self
        ctx["env"] = self.env

        def render(path, context=None, renderer="text"):
            if isinstance(context, dict):
                context = {**ctx, **context}
            else:
                context = ctx

            if "basedir" in context:
                path = os.path.join(context["basedir"], path)

            file = os.path.abspath(path)

            if not os.path.isfile(file):
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), file)

            return self._compile(file, renderer, None, None, context)

        ctx["render"] = render

        kwargs["tower"] = ctx["tower"]
        kwargs["minion_id"] = ctx["minion_id"]

        # Explicitly set `saltenv=None` to disable any special behavior
        # of some rendering engines (mostly JINJA) regarding rendering
        # states or pillars.
        #
        # Salt does not provide an option to pass any file lookup
        # customization to the renderers. Therefore some renderering
        # engines strictly assume that if `saltenv` is not `None` they
        # are rendering salt states, and if the internal variable
        # `_pillar_rend` is set, they are rendering pillars. They will
        # then use a different file lookup mechanism using the
        # configured pillar_roots or file_roots. This cannot be
        # customized.
        #
        # It is currently impossible to lookup files relative to the
        # tower.sls directory (which might or might not be also a
        # pillar_roots) for these renderers without explicitly using a
        # helper function.
        kwargs["saltenv"] = None
        kwargs["_pillar_rend"] = True

        if __opts__.get("salt_tower.unstable_enable_saltenv", False):
            kwargs["saltenv"] = self.env

        try:
            return salt.template.compile_template(
                template=template,
                renderers=self._renderers,
                default=default,
                blacklist=blacklist,
                whitelist=whitelist,
                context=ctx,
                **kwargs,
            )
        except Exception as err:
            LOGGER.critical("Unable to render template `%s'", template)
            LOGGER.exception(err)
            raise err


class Formatter(string.Formatter):
    def __init__(self, tower):
        self._tower = tower

    def get_field(self, field_name, args, kwargs):
        if field_name in kwargs:
            return (kwargs[field_name], None)

        value = self._tower.get(field_name, delimiter=".")

        if value is None:
            return ("{" + field_name + "}", None)

        return (value, None)


def _merge(tgt, *objects, strategy="merge-last"):
    for obj in objects:
        if isinstance(tgt, dict):
            tgt = _merge_dict(tgt, obj, strategy)
        elif isinstance(tgt, list):
            tgt = _merge_list(tgt, obj, strategy)
        else:
            raise TypeError(f"Cannot merge {type(tgt)}")

    return tgt


def _merge_clean(obj):
    if isinstance(obj, dict):
        return {k: _merge_clean(v) for k, v in obj.items() if k != "__"}

    if isinstance(obj, list):
        if obj and isinstance(obj[0], dict) and len(obj[0]) == 1 and "__" in obj[0]:
            obj = obj[1:]
        return [_merge_clean(v) for v in obj]

    return obj


def _merge_dict(tgt, obj, strategy="merge-last"):  # pylint: disable=too-many-branches
    if not isinstance(obj, dict):
        raise TypeError(f"Cannot merge non-dict type, but is {type(obj)}")

    if "__" in obj:
        strategy = obj.pop("__")

    if strategy == "remove":
        for k in obj:
            if k in tgt:
                tgt.pop(k)

    elif strategy == "merge-last":
        for key, val in obj.items():
            if key in tgt and isinstance(tgt[key], dict) and isinstance(val, dict):
                _merge_dict(tgt[key], val, strategy)
            elif key in tgt and isinstance(tgt[key], list) and isinstance(val, list):
                _merge_list(tgt[key], val, strategy)
            else:
                tgt[key] = _merge_clean(val)

    elif strategy == "merge-first":
        for key, val in obj.items():
            if key in tgt and isinstance(tgt[key], dict) and isinstance(val, dict):
                _merge_dict(tgt[key], val, strategy)
            elif key in tgt and isinstance(tgt[key], list) and isinstance(val, list):
                _merge_list(tgt[key], val, strategy)
            elif key not in tgt:
                tgt[key] = _merge_clean(val)

    elif strategy == "overwrite":
        tgt.clear()
        tgt.update(_merge_clean(obj))

    else:
        raise ValueError(f"Unknown strategy: {strategy}")

    return tgt


def _merge_list(tgt, lst, strategy="merge-last"):
    if not isinstance(lst, list):
        raise TypeError(f"Cannot merge non-list type, but is {type(lst)}")

    if lst and isinstance(lst[0], dict) and len(lst[0]) == 1 and "__" in lst[0]:
        strategy = lst.pop(0)["__"]

    if strategy == "remove":
        for val in lst:
            if val in tgt:
                tgt.remove(val)

    elif strategy == "merge-last":
        tgt.extend(_merge_clean(lst))

    elif strategy == "merge-first":
        for val in _merge_clean(lst):
            tgt.insert(0, val)

    elif strategy == "overwrite":
        del tgt[:]
        tgt.extend(_merge_clean(lst))

    else:
        raise ValueError(f"Unknown strategy: {strategy}")

    return tgt



# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring
# pylint: disable=redefined-outer-name


def test_render(render):
    template = """
        A small text
    """

    assert render(template, default="text") == "A small text\n"


def test_strip(render):
    template = """
        #!text strip

            Indented text

        END

    """

    assert render(template) == "Indented text\n\nEND"


def test_key(render):
    template = """
        #!text key=a:b:c
        text value
    """

    assert render(template) == {"a": {"b": {"c": "text value\n"}}}



# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring
# pylint: disable=redefined-outer-name

from __future__ import absolute_import

import os
import textwrap

import pytest
import salt.config
import salt.loader

try:
    import salt.features

    SALT_FEATURES_AVAILABLE = True
except ImportError:
    SALT_FEATURES_AVAILABLE = False


ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))


@pytest.fixture
def env(tmpdir):
    return Environment(tmpdir)


@pytest.fixture
def features():
    return {
        # This feature will become the default in the Phoshorus release.
        "enable_slsvars_fixes": True,
    }


@pytest.fixture(autouse=SALT_FEATURES_AVAILABLE)
def setup_features(features):
    # pylint: disable=no-member
    salt.features.features.setup = True
    salt.features.features.features = features


@pytest.fixture
def render(env):
    return env.render


class Environment:
    def __init__(self, tmpd):
        self.tmpd = tmpd

        self.opts = salt.config.client_config(os.path.join(ROOT, "test/master.yml"))
        self.opts["cachedir"] = os.path.join(tmpd, "cache")

        if os.getenv("USE_PACKAGE", "") == "":
            self.opts["module_dirs"] = [ROOT]

        self.opts["pillar_roots"] = {
            "base": [os.path.join(ROOT, "test/fixtures/pillar")]
        }

        self.pillar = {}
        self.grains = {"id": self.opts["id"]}
        self.opts["grains"] = self.grains
        self.opts["pillar"] = self.pillar

    def setup(self, files):
        for k, v in files.items():
            self.write(k, textwrap.dedent(v).strip())

    def write(self, name, content, mode="w"):
        template = self.tmpd / name
        template.write(content, mode, ensure=True)

    def build(self):
        return Master(self.tmpd, self.opts)

    def compile_template(self, *args, **kwargs):
        return self.build().compile_template(*args, **kwargs)

    def ext_pillar(self, *args, **kwargs):
        return self.build().ext_pillar(*args, **kwargs)

    def render(self, *args, **kwargs):
        return self.build().render(*args, **kwargs)


class Master:
    def __init__(self, tmpd, opts):
        self.tmpd = tmpd
        self.opts = opts
        self.minion_mods = salt.loader.minion_mods(self.opts)
        self.renderers = salt.loader.render(self.opts, self.minion_mods)
        self.pillars = salt.loader.pillars(self.opts, self.minion_mods)

    def compile_template(self, template, default="jinja|yaml", **kwargs):
        template = os.path.join(self.tmpd, template)
        return salt.template.compile_template(
            template,
            self.renderers,
            default,
            self.opts["renderer_blacklist"],
            self.opts["renderer_whitelist"],
            **kwargs,
        )

    def ext_pillar(self, **kwargs):
        args = kwargs.pop("args", [os.path.join(self.tmpd, "tower.sls")])
        pillar = self.opts["pillar"]
        minion_id = kwargs.pop("minion_id", self.opts["id"])

        if isinstance(args, dict):
            return self.pillars["tower"](minion_id, pillar, **args)
        if isinstance(args, list):
            return self.pillars["tower"](minion_id, pillar, *args)
        return self.pillars["tower"](minion_id, pillar, args)

    def render(self, content, mode="w", **kwargs):
        path = self.tmpd / "unnamed.sls"
        path.write(textwrap.dedent(content).lstrip(), mode, ensure=True)
        return self.compile_template(path, **kwargs)
